#!/usr/bin/python3

# Copyright (c) 2025 ModalAI Inc.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# 4. The Software is used solely in conjunction with devices provided by
#    ModalAI Inc.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# For a license to use on non-ModalAI hardware, please contact license@modalai.com

from voxl_esc_setup_paths import *
voxl_esc_setup_paths()

from libesc import *
from esc_scanner import EscScanner
import argparse

parser = argparse.ArgumentParser(description='ESC Params Verification Script')
parser.add_argument('--device',        required=False, default=None)
parser.add_argument('--baud-rate',     required=False, default=None)
parser.add_argument('--num-escs',      type=int, required=False,  default=4)
parser.add_argument('--save-params',   type=int, required=False,  default=0)
parser.add_argument('--params-search-path', type=str, required=False, default='../voxl-esc-params')
parser.add_argument('--verify-exact-match', type=str, required=False, default='')
args = parser.parse_args()

devpath     = args.device
baudrate    = args.baud_rate
num_escs    = args.num_escs
save_params = args.save_params
params_search_path = args.params_search_path
exact_match_param_file = args.verify_exact_match

# quick scan for ESCs to detect the port
scanner = EscScanner()
(devpath, baudrate) = scanner.scan(devpath, baudrate)

if devpath is not None and baudrate is not None:
    print('INFO: ESC(s) detected on port: ' + devpath + ', baud rate: ' + str(baudrate))
else:
    print('ERROR: No ESC(s) detected, exiting.')
    sys.exit(1)

try:
    esc_manager = EscManager()
    esc_manager.open(devpath, baudrate)
except Exception as e:
    print('ERROR: Unable to open serial port :')
    print(e)
    sys.exit(1)

# wait a little to let manager find all ESCs and get info
time.sleep(0.1)

escs = esc_manager.get_escs()
esc_ids = [e.get_id() for e in escs]

num_invalid_params = 0
num_exact_match    = 0

def wait_for_params(e, timeout, param_type):
    t_start = time.time()
    while (time.time()-t_start < timeout):
        if param_type == 'board' and e.params.is_valid_board():
            return True
        if param_type == 'id' and e.params.is_valid_id():
            return True
        if param_type == 'uart' and e.params.is_valid_uart():
            return True
        if param_type == 'tune' and e.params.is_valid_tune():
            return True
            
        time.sleep(0.005)

    return False

# send request for each ESC to send back its params
for esc_id in range(num_escs):
    if esc_id not in esc_ids:
        print('ERROR: ESC ID %d not found' % (esc_id))
        continue

    esc = esc_manager.get_esc_by_id(esc_id)

    param_timeout = 0.200
    esc_manager.request_config_id(esc_id)
    wait_for_params(esc,param_timeout,'id');
    esc_manager.request_config_board(esc_id)
    wait_for_params(esc,param_timeout,'board');
    esc_manager.request_config_uart(esc_id)
    wait_for_params(esc,param_timeout,'uart');
    esc_manager.request_config_tune(esc_id)
    wait_for_params(esc,param_timeout,'tune');


    if not esc.params.is_valid():
        print('ERROR: Params for ID %d are invalid!' % (esc_id))
        num_invalid_params +=1
    else:
        # get the params as XML string and save to a file
        params_str = esc.params.get_xml_string()

        if not exact_match_param_file:
            # try to find a match with existing xml file in params directory
            params_candidates = glob.glob(params_search_path + '/**/*.xml', recursive=True)
        else:
            params_candidates = [exact_match_param_file]

        for param_file in params_candidates:
            try:
                with open(param_file, 'r') as f:
                    temp_param_str = f.read()
                try_params = params_from_xml(temp_param_str)
                if try_params.get_param_bytes_all() == esc.params.get_param_bytes_all():
                    print('INFO: Params from ID %d match %s' % (esc_id,param_file))
                    num_exact_match += 1
            except Exception as e:
                print('ERROR: Unable to open param file %s' % (param_file))
                print(e)

        if save_params:
            xml_out_file = 'esc%d_params.xml' % (esc_id)
            with open(xml_out_file , 'w') as f:
                f.write(params_str)
                print('INFO: Saved params from ESC ID %d to %s' %(esc_id, xml_out_file))

#compare all params from all ESCs
num_params_match = 0;
for esc_id in esc_ids:
    if esc_manager.get_esc_by_id(esc_ids[0]).params.get_param_bytes_all() != \
        esc_manager.get_esc_by_id(esc_id).params.get_param_bytes_all():
        print('ERROR: Params from ID %d and %d are not the same' % (esc_ids[0],esc_id))
    else:
        num_params_match += 1

esc_manager.close()

if exact_match_param_file:
    if num_exact_match != num_escs:
        print('Number of exact param matches %d is not equal to %d!' %(num_exact_match, num_escs))
        sys.exit(1)
    else:
        print('All %d ESC params matched to %s' %(num_escs,exact_match_param_file))

if num_escs == len(esc_ids) and num_params_match == num_escs and num_invalid_params == 0:
    print('INFO: Success! Params in all ESCs are valid and identical.')
else:
    print('ERROR: Some params are invalid or not the same!')
    print('       Number of ESCs expected     : %d' % num_escs)
    print('       Number of ESCs found        : %d' % (len(esc_ids)))
    print('       Number of invalid params    : %d' % (num_invalid_params))
    print('       Number of matched params    : %d' % (num_params_match))
    sys.exit(1)
